@@ -222,7 +222,10 @@ def get_subnet_id(name) -> str:
     return _resolve_id('subnets', name)
 
 
-def create_subnet(subnet_name, network_id, cidr='192.168.1.0/24',
+def create_subnet(subnet_name, network_id,
+                  cidr='192.168.1.0/24',
+                  allocation_pool_start=None,
+                  allocation_pool_end=None,
                   gateway_ip=None) -> dict:
     """Create a subnet on a network.
 
@@ -245,19 +248,20 @@ def create_subnet(subnet_name, network_id, cidr='192.168.1.0/24',
     }
     if gateway_ip:
         subnet['gateway_ip'] = gateway_ip
-
-    subnet = neutron().create_subnet(body={
-        'subnets': [
+    if allocation_pool_start and allocation_pool_end:
+        subnet['allocation_pools']= [
             {
-                'name': subnet_name,
-                'cidr': cidr,
-                'ip_version': 4,
-                'network_id': network_id,
+            "start": allocation_pool_start,
+            "end": allocation_pool_end,
             }
         ]
+
+
+    subnet_rtn = neutron().create_subnet(body={
+        'subnets': [ subnet ]
     })
 
-    return subnet['subnets'][0]
+    return subnet_rtn['subnets'][0]
 
 
 def delete_subnet(subnet_id):
@@ -609,19 +613,31 @@ def remove_subnet_from_router(router_id, subnet_id):
 # Floating IPs
 ###############
 
-def get_free_floating_ip() -> dict:
+def get_free_floating_ip(allocate=True) -> dict:
     """Get the first unallocated floating IP available to your project.
 
+    Args:
+        allocate (bool): Whether to allocate a new floating IP if there are no
+            Floating IPs currently free in your project. Defaults to True.
+
     Returns:
         The free floating IP representation.
     """
-    ips = neutron().list_floatingips()['floatingips']
+    _neutron = neutron()
+    ips = _neutron.list_floatingips()['floatingips']
     unbound = (ip for ip in ips if ip['port_id'] is None)
     try:
         fip = next(unbound)
         return fip
     except StopIteration:
-        raise Exception("No free floating IP found")
+        if not allocate:
+            raise RuntimeError(
+                "No free floating IPs in project and not allocating a new one")
+        return _neutron.create_floatingip({
+            "floatingip": {
+                "floating_network_id": get_network_id(PUBLIC_NETWORK),
+            }
+        })["floatingip"]
 
 
 def get_or_create_floating_ip() -> 'tuple[dict,bool]':
@@ -636,7 +652,8 @@ def get_or_create_floating_ip() -> 'tuple[dict,bool]':
             available to allocate.
     """
     try:
-        fip = get_free_floating_ip()
+        # Set allocate=False so we explicitly know if we created one.
+        fip = get_free_floating_ip(allocate=False)
         created = False
     except Exception:
         network_id = get_network_id(PUBLIC_NETWORK)
