@@ -1,7 +1,10 @@
 from datetime import datetime
 from operator import attrgetter
+import socket
+import time
 
 from novaclient.v2.flavor_access import FlavorAccess as NovaFlavor
+from novaclient.v2.keypairs import Keypair as NovaKeypair
 from novaclient.v2.servers import Server as NovaServer
 from novaclient.exceptions import NotFound
 
@@ -11,13 +14,14 @@ from .keypair import Keypair
 from .image import get_image, get_image_id
 from .network import (get_network_id, get_or_create_floating_ip,
                       get_floating_ip, get_free_floating_ip)
-from .util import random_base32
+from .util import random_base32, sshkey_fingerprint
 
 __all__ = [
     'get_flavor',
     'get_flavor_id',
     'show_flavor',
     'show_flavor_by_name',
+    'list_flavors',
 
     'get_server',
     'get_server_id',
@@ -27,6 +31,14 @@ __all__ = [
     'show_server_by_name',
 
     'create_server',
+
+    'associate_floating_ip',
+    'detach_floating_ip',
+
+    'wait_for_active',
+    'wait_for_tcp',
+
+    'update_keypair',
 ]
 
 DEFAULT_IMAGE = 'CC-CentOS7'
@@ -334,6 +346,15 @@ def show_flavor_by_name(name) -> NovaFlavor:
     return show_flavor(flavor_id)
 
 
+def list_flavors() -> 'list[NovaFlavor]':
+    """Get a list of all available flavors.
+
+    Returns:
+        A list of all flavors.
+    """
+    return nova().flavors.list()
+
+
 ##########
 # Servers
 ##########
@@ -427,34 +448,149 @@ def show_server_by_name(name) -> NovaServer:
     return show_server(server_id)
 
 
-def associate_floating_ip(server_name, floating_ip_address=None):
-    server_id = get_server_id(server_name)
+def associate_floating_ip(server_id, floating_ip_address=None):
+    """Associate an allocated Floating IP with a server.
 
-    if floating_ip_address:
-        fip = get_floating_ip(floating_ip_address)
-    else:
-        fip = get_free_floating_ip()
+    If no Floating IP is specified, one will be allocated dynamically.
 
-    ip = fip['floating_ip_address']
-    connection().compute.add_floating_ip_to_server(server_id, ip)
+    Args:
+        server_id (str): The ID of the server.
+        floating_ip_address (str): The IPv4 address of the Floating IP to
+            assign. If specified, this Floating IP must already be allocated
+            to the project.
 
-    return ip
+    """
+    if not floating_ip_address:
+        floating_ip_address = get_free_floating_ip()["floating_ip_address"]
+
+    connection().compute.add_floating_ip_to_server(server_id, floating_ip_address)
+
+    return floating_ip_address
 
 
-def detach_floating_ip(server_name, floating_ip_address):
-    server_id = get_server_id(server_name)
+def detach_floating_ip(server_id, floating_ip_address):
+    """Remove an allocated Floating IP from a server by name.
+
+    Args:
+        server_id (str): The name of the server.
+        floating_ip_address (str): The IPv4 address of the Floating IP to
+            remove from the server.
+
+    """
     connection().compute.remove_floating_ip_from_server(
         server_id, floating_ip_address)
 
 
+def wait_for_active(server_id, timeout=(60 * 20)):
+    """Wait for the server to go in to the ACTIVE state.
+
+    If the server goes in to an ERROR state, this function will terminate. This
+    is a blocking function.
+
+    .. note::
+
+       For bare metal servers, when the server transitions to ACTIVE state, this
+       actually indicates it has started its final boot. It may still take some
+       time for the boot to complete and interfaces e.g., SSH to come up.
+
+       If you want to wait for a TCP service like SSH, refer to
+       :func:`wait_for_port`.
+
+    Args:
+        server_id (str): The ID of the server.
+        timeout (int): The number of seconds to wait for before giving up.
+            Defaults to 20 minutes.
+
+    """
+    from openstack.compute.v2.server import Server
+    compute = connection().compute
+    server = compute._get_resource(Server, server_id)
+    return compute.wait_for_server(server, wait=timeout)
+
+
+def wait_for_tcp(host, port, timeout=(60 * 20)):
+    """Wait until a port on a server starts accepting TCP connections.
+
+    The implementation is taken from `wait_for_tcp_port.py
+    <https://gist.github.com/butla/2d9a4c0f35ea47b7452156c96a4e7b12>`_.
+
+    Args:
+        host (str): The host that should be accepting connections. This can
+            be either a Floating IP or a hostname.
+        port (int): Port number.
+        timeout (int): How long to wait before raising errors, in seconds.
+            Defaults to 20 minutes.
+
+    Raises:
+        TimeoutError: If the port isn't accepting connection after time
+            specified in `timeout`.
+    """
+    start_time = time.perf_counter()
+
+    while True:
+        try:
+            with socket.create_connection((host, port), timeout=timeout):
+                break
+        except OSError as ex:
+            time.sleep(0.01)
+            if time.perf_counter() - start_time >= timeout:
+                raise TimeoutError((
+                    f'Waited too long for the port {port} on host {host} to '
+                    'start accepting connections.')) from ex
+
+############
+# Key pairs
+############
+
+def update_keypair(key_name=None, public_key=None) -> "NovaKeypair":
+    """Update a key pair's public key.
+
+    Due to how OpenStack Nova works, this requires deleting and re-creating the
+    key even for public key updates. The key will not be re-created if it
+    already exists and the fingerprints match.
+
+    Args:
+        key_name (str): The name of the key pair to update. Defaults to value
+            of the "key_name" context variable.
+        public_key (str): The public key to update the key pair to reference.
+            Defaults to the contents of the file specified by the
+            "keypair_public_key" context variable.
+
+    Returns:
+        The updated (or created) key pair.
+    """
+    if not key_name:
+        key_name = get_from_context("key_name")
+    if not public_key:
+        public_key_path = get_from_context("keypair_public_key")
+        if public_key_path:
+            with open(public_key_path, "r") as pubkey:
+                public_key = pubkey.read().strip()
+
+    assert key_name is not None
+    assert public_key is not None
+
+    _nova = nova()
+    try:
+        existing = _nova.keypairs.get(key_name)
+        if existing.fingerprint == sshkey_fingerprint(public_key):
+            return existing
+        _nova.keypairs.delete(key_name)
+        return _nova.keypairs.create(
+            key_name, public_key=public_key, key_type="ssh")
+    except NotFound:
+        return _nova.keypairs.create(
+            key_name, public_key=public_key, key_type="ssh")
+
+
 ##########
 # Wizards
 ##########
 
-def create_server(server_name, reservation_id, key_name=None, network_id=None,
+def create_server(server_name, reservation_id=None, key_name=None, network_id=None,
                   network_name=DEFAULT_NETWORK, nics=[], image_id=None,
                   image_name=DEFAULT_IMAGE, flavor_id=None,
-                  flavor_name=BAREMETAL_FLAVOR, count=1) -> 'Union[NovaServer,list[NovaServer]]':
+                  flavor_name=None, count=1) -> 'Union[NovaServer,list[NovaServer]]':
     """Launch a new server instance.
 
     Args:
@@ -464,7 +600,8 @@ def create_server(server_name, reservation_id, key_name=None, network_id=None,
             reservation for bare metal server instances.
         key_name (str): A key pair name to associate with the server. Any user
             holding the private key for the key pair will be able to SSH to
-            the instance as the ``cc`` user.
+            the instance as the ``cc`` user. Defaults to the key specified
+            by the "key_name" context variable.
         network_id (str): The network ID to connect the server to. The server
             will obtain an IP address on this network when it boots.
         network_name (str): The name of the network to connect the server to.
@@ -474,10 +611,11 @@ def create_server(server_name, reservation_id, key_name=None, network_id=None,
         image_name (str): The name of the image to user for the server's disk
             image. If ``image_id`` is also set, that takes priority.
             (Default ``DEFAULT_IMAGE``.)
-        flavor_id (str): The flavor ID to use when launching the server.
-            (Default ``BAREMETAL_FLAVOR``).
+        flavor_id (str): The flavor ID to use when launching the server. If not
+            set, and no ``flavor_name`` is set, the first flavor found is used.
         flavor_name (str): The name of the flavor to use when launching the
-            server. If ``flavor_id`` is also set, that takes priority.
+            server. If ``flavor_id`` is also set, that takes priority. If not
+            set, and no ``flavor_id`` is set, the first flavor found is used.
         count (int): The number of instances to launch. When launching bare
             metal server instances, this number must be less than or equal to
             the total number of hosts reserved. (Default 1).
@@ -492,7 +630,7 @@ def create_server(server_name, reservation_id, key_name=None, network_id=None,
     if count < 1:
         raise ValueError('Must launch at least one server.')
     if not key_name:
-        key_name = get_from_context('key_name')
+        key_name = update_keypair().id
     if not network_id:
         network_id = get_network_id(network_name)
     if not nics:
@@ -500,12 +638,22 @@ def create_server(server_name, reservation_id, key_name=None, network_id=None,
     if not image_id:
         image_id = get_image_id(image_name)
     if not flavor_id:
-        flavor_id = get_flavor_id(flavor_name)
+        if flavor_name:
+            flavor_id = get_flavor_id(flavor_name)
+        else:
+            flavor_id = next((f.id for f in list_flavors()), None)
+            if not flavor_id:
+                raise NotFound('Could not auto-select flavor to use')
+
+    scheduler_hints = {}
+    if reservation_id:
+        scheduler_hints['reservation'] = reservation_id
+
     server = nova().servers.create(
         name=server_name,
         image=image_id,
         flavor=flavor_id,
-        scheduler_hints={'reservation': reservation_id},
+        scheduler_hints=scheduler_hints,
         key_name=key_name,
         nics=nics,
         min_count=count,
